#!/usr/bin/env python

'''
Subroutines related to creating and manipulating rfl configuration files.

Mike Markowski, mike.ab3ap@gmail.com
Apr 2023
'''

import numpy as np
import sys

def cfgDefault():
	'''Create a default rfl configuration file that is fully commented.

	Input:
	  none
	Output:
	  (string): configuration file as a large string.
	'''

	return '''
# There are 2 styles of output: coverage map or link data.
# There are 2 styles of calculation: profile and area.
#
# The result is 4 possible combinations.  A coverage map is used to visualize
# how a transmitter deposits energy on a map, whereas link data calculates
# results for all pairwise tx-rx links.  Multiple transmitters and multiple
# receivers can exist, but tx-tx and rx-rx are not calculated.  In coverage
# mode, only the first transmitter found is used.  If multiple receivers
# exist, environmental and antenna data are used from the first receiver
# found.

# Items particular to rfl itself.
model rfl
  terrain dted        # 'dted' or 'open', indicating type of elevation data.
  dted /data/dted1    # Path to DTED1 data.
  ant  /data/ant      # Path to NSMA antenna patterns.
  mode area           # ITM analysis type, 'profile' or 'area'.
  dh 127              # Terrain roughness, only needed in area mode.
  output coverage     # 'coverage' (graph) or 'csv' (file of numbers).
  outfile rfl.png     # Output csv file name, not needed in coverage mode.
  units dBuV          # Output units: dBm, uV, dBuV, SNR.

  # Statistics involve how, where and when radios are used.  Rather than mean
  # and standard deviation ITM uses quantiles of observations, values not
  # exceeded for given fractions of measure.  Quantiles of time, location
  # and situation are used, which are combined in different ways to apply to
  # what ITM calls modes of variability.  For detail, see See "A Guide to the
  # ITS Irregular Terrain Model in the Area Prediction Mode," Hufford,
  # Longley and Kissick, 1982, NTIA 82-100.
  #
  # ITM modes of variability (mdvar variable in itm.py):
  # Single message probability: first-try success probability.
  # Single receiver: % of time availibility.
  # Mobile receiver: % of required reliability.
  # Broadcast service: % of time that % locations have reception,
  #
  # Single message mode: no variability in time, location or situation.
  # Profile mode: no variability in location.
  # Mobile mode: no variability in situation.
  # Broadcast mode: variability in all three quantiles.
  #
  # Reliability is a quantile that addresses adequate service (or measures
  # mentioned above) capturing variability of a radio system during the
  # course of its deployment.  Confidence in ITM is a higher level of
  # probability where a large number of engineering decisions at the same
  # confidence level will be correct for that fraction of decisions.
  #
  # If the quantiles are not in the config file then rfl will not calculate
  # the statistics.

  conf 50           # 0-100, confidence level.
  time 50           # 0-100, quantile of time.
  loc  50           # 0-100, quantile of location.

# If coverage block does not exist, point to point tx-rx
# link budgets will be calculated.
model coverage
  grid  20        # Make coverage map n x n cells.
  range 80        # Range of coverage in km from tx point.
  title W3GMS/R Coverage
  levels -14,6,26,46,66,86,106,126 # Isoline levels.
  lat   40.008390 # Map center, decimal degrees (optional).
  lon   -75.93459 # Map center, decimal degrees (optional).
  pattern yes     # Display NSMA pattern, yes/no.
  contour line    # 'line' or 'fill' for isolevel lines or splotches.

model env easternPA

  # Climate index is one of:
  #   0   Equatorial
  #   1   Continental subtropical
  #   2   Maritime subtropical
  #   3   Desert
  #   4   Continental temperate
  #   5   Maritime temperate, overland
  #   6   Maritime temperate, oversea
  climate 4
  eps   15    # average relative ground permittivity, F/m. (LR1968, tbl 2.)
  N0    315   # sea level refractivity, mid-Atlantic (TN101 fig 4.1).
  sigma 0.005 # average relative ground conductivity, Siemens. (LR1968, tbl 2.)

model noise suburban
  B    25e3        # Hz, bandwidth of signal.
  env  residential # string, environment name.
  f    146.985     # MHz, carrier frequency.
  NF   7           # dB, receiver noise figure.
  T    288         # Kelvin, temperature of receiver.

# Multiple stations can exist.  Coverage mode uses only the first transmitter.
model station W3GMS/R
  # Antenna information.
  ant   8.15          # NSMA file or gain in dBi.
  pol   v             # 'v' or 'h', antenna polarity.
  az    0             # Degrees clockwise from true north.
  el    0             # Degrees above horizontal.

  # Station info.
  hgt   24.4        # meters above earth surface.
  lat   40.008390   # decimal degrees.  Optional in area mode.
  lon   -75.93459   # decimal degrees.  Optional in area mode.
  env   easternPA   # Name of environment model in this file.
  noise suburban    # Name of noise model in this file.
  sys   12          # dB, various system losses, cabling, insertion, etc.
  tx    54.77       # dBm.
  site  v           # Siting r(andom), c(areful), v(ery careful) in area mode.
  # For transmitters, transmission service (mode of varability) index:
  #   0   Single message mode.  Time, location, and situation
  #       variability are combined together to give a con-
  #       fidence level
  #   1   Individual broadcast receiver mode.  Reliability is given by time
  #       availability for the individuial receiver of a broadcast station.
  #       Confidence is a combination of location and situation variability.
  #   2   Mobile mode.  Reliability is a combination of time
  #       and location variability.  Confidence is given by the
  #       situation variability.
  #   3   Broadcast mode.  Reliability is given by the twofold
  #       statement of -at least- qT of the time in qL of the
  #       locations.  Confidence is given by the situation
  #       variability.
  service 1

model station AB3AP # Me!
  env   easternPA
  hgt   4
  noise suburban
  sys   15          # dB, various system losses, cabling, insertion, etc.
'''

def cfgEmpty():
	'''Create an empty rfl configuration file for the GUI.

	Input:
	  none
	Output:
	  (dictionary): an empty rfl config dictionary.
	'''

	d = {}
	d['rfl'] = {}
	d['rfl']['ant'] = ''
	d['rfl']['terrain'] = ''
	d['rfl']['dted'] = ''
	d['rfl']['outfile'] = ''
	d['rfl']['conf'] = ''
	d['rfl']['time'] = ''
	d['rfl']['loc'] = ''
	d['rfl']['dh'] = ''
	d['rfl']['interference'] = 'no' # Not an interference study.

	d['coverage'] = {}
	d['coverage']['grid'] = ''
	d['coverage']['range'] = ''
	d['coverage']['title'] = ''
	d['coverage']['levels'] = ''

	d['env'] = {}
	d['noise'] = {}
	d['station'] = {}
	return d

def cfgEmptyEnv():
	e = {}
	e['eps'] = 0
	e['sigma'] = 0
	e['N0'] = 0
	e['climate'] = 4 # Continental temperate.
	return e

def cfgEmptyNoise():
	n = {}
	n['B'] = 0
	n['env'] = ''
	n['f'] = '0'
	n['NF'] = 0
	n['K'] = 0
	return n

def cfgEmptyStation():
	s = {}
	s['ant'] = ''
	s['pol'] = ''
	s['az'] = 0
	s['el'] = 0
	s['hgt'] = 0
	s['lat'] = 0
	s['lon'] = 0
	s['env'] = ''
	s['noise'] = ''
	s['service'] = 3 # Broadcast service.
	s['site'] = ''
	s['sys'] = 0
	return s

def cfgParse(cfgData):
	'''Read an rfl configuration file and return its contents in a string.

	Input:
	  filename (string): file name of rfl configuration file.
	Output:
	  (string): config file contents.
	'''

	d = {}
	d['coverage'] = {}
	d['coverage']['contour'] = 'fill'
	d['coverage']['pattern'] = 'no'
	d['env'] = {}
	d['noise'] = {}
	d['rfl'] = {}
	d['rfl']['interference'] = 'no'
	d['rfl']['conf'] = d['rfl']['time'] = d['rfl']['loc'] = 50
	d['station'] = {}

	lineno = 0
	for line in cfgData: # Loop through file line by line.
		lineno += 1 # Line count for error messages.

		# Ignore '#' comments and blank lines.
		excl = line.find('#')
		if excl != -1: # Ignore comments.
			line = line[:excl].strip()
		line = line.strip()
		if line == '': # Ignore blank lines.
			continue

		field = line.split()
		if field[0] == 'model': # Start of 'model' blocks.
			match field[1]:
				case 'coverage':
					d2 = d['coverage']
					d2['title'] = ''
				case 'env':
					curE = field[2]
					d2 = d['env'][curE] = defEnv()
				case 'noise':
					curN = field[2]
					d2 = d['noise'][curN] = defNoise()
				case 'rfl':
					d2 = d['rfl']
				case 'station':
					letterN = line.find('n')
					curS = line[2+letterN:].strip()
					d2 = d['station'][curS] = defStation()
		else: # Fields within model block.
			if field[0] in ['title']:
				letterE = line.find('e')
				val = line[2+letterE:].strip()
			elif field[0] in ['ant','contour','dted','env','f','interference',
				'mode','noise', 'outfile','output','pattern','pol','site',
				'terrain','units']:
				val = field[1]                             # String.
			elif field[0] in ['climate', 'grid', 'service']:
				val = int(field[1])                        # Integer
			elif field[0] in ['levels']:
				val = list(map(float,field[1].split(','))) # List of floats.
			elif field[0] in ['az','el','lat','lon']:
				val = np.radians(float(field[1]))          # Radians.
			else:
				val = float(field[1])                      # Float.
				if field[0] == 'tx' and 'service' not in d2.keys():
					d2['service'] = 3 # Hack, default service for tx.
			d2[field[0]] = val
	return d # Dictionary containing config file.

def cfgRead(filename):
	'''Read an rfl configuration file and return its contents in a python
	dictionary.

	Input:
	  filename (string): file name of rfl configuration file.
	Output:
	  (dictionary): dictionary representation of contents of config file.
	'''

	try:
		f = open(filename)
	except FileNotFoundError:
		print('rfl: %s not found.' % filename)
		sys.exit(1)
	cfgData = f.read().split('\n') # Read configuration file.
	f.close()
	return cfgData

def cfgWrite(filename):
	'''Create a file containing a default rfl configuration file.

	Input:
	  filename (string): name of rfl configuration file to create.
	Output:
	  none
	'''

	f = open(filename, 'w')
	f.write(cfgDefault())
	f.close()

def defEnv():
	'''Create a defalt (eastern Pennsylvania) EM environment dictionary.'''
	e = {}
	e['eps'] = 10
	e['N0'] = 315
	e['sigma'] = 0.005
	e['climate'] = 4 # Continental temperate.
	return e

def defNoise():
	'''Create a defalt (eastern Pennsylvania!) noise environment dictionary.'''
	n = {}
	n['B'] = 25e3
	n['env'] = 'residential'
	n['f'] = '146.985'
	n['NF'] = 7
	n['T'] = 288
	return n

def defStation():
	'''Create a defalt station dictionary.'''
	s = {}
	s['ant'] = '0'
	s['az'] = 0
	s['el'] = 0
	s['hgt'] = 1
	s['pol'] = 'v'
	s['site'] = 'r'
	s['sys'] = 0
	return s

