# Commonly used functions.
from numpy import arccos, arcsin, arctan, arctan2, cos, pi, sin, sqrt

import numpy as np
import numpy.linalg as la

'''
Natl Geodetic Survey code to calculate various values associated with
WGS84 model of the earth. Converted original Fortran code as retrieved
from http://www.ngs.noaa.gov/TOOLS/index.shtml to python.

Used in parts of several subroutine names, it seems that GPN means
Geographic PositioN.

The methods dirct1, gpnhri, and getrad were converted from Fortran. The
Fortran source was retrieved from http://www.ngs.noaa.gov/TOOLS/index.shtml
at the sub-page http://www.ngs.noaa.gov/TOOLS/program_descriptions.html/InvFwd

WGS 84 info:
http://earth-info.nga.mil/GandG/publications/tr8350.2/wgs84fin.pdf, NIMA
Technical Report TR8350.2, Department of Defense World Geodetic System 1984,
Its Definition and Relationships with Local Geodetic Systems (especially
chapter 3).  http://earth-info.nga.mil/GandG/publications/historic/historic.html,
DMA Technical Report TR8350.2-A, Supplement To Department of Defense World
Geodetic System 1984 Technical Report Methods, Techniques, and Data Used in
WGS&nbsp84 Development (especially
http://earth-info.nga.mil/GandG/publications/tr8350.2/tr8350.2-a/Chapter%203.pdf
chapter 3.

Mike Markowski, mike.ab3ap@gmail.com
Dec 2022
'''

aMax_m = 6378137           # m, WGS84 semimajor axis equatorial.
bMax_m = 6356752.314245179 # m, WGS84 semiminor, polar.
#fRec = 298.25722210088  # GRS-80 value. fRec = 1/f = a/(a - b)
fRec = 298.257223563     # WGS84 value, unitless..
flat = 1/fRec            # Ellipsoid flattening, unitless.
eSq = flat*(2 - flat)    # Eccentricity squared, WGS84 model.

def direct(lat_rad, lon_rad, fAz_rad, s_m): # Easier name.
    return dirct1(lat_rad, lon_rad, fAz_rad, s_m)

def dirct1(lat1_rad, lon1_rad, fAz_rad, s_m):

    '''
    C
    C *** SOLUTION OF THE GEODETIC DIRECT PROBLEM AFTER T.VINCENTY
    C *** MODIFIED RAINSFORD'S METHOD WITH HELMERT'S ELLIPTICAL TERMS
    C *** EFFECTIVE IN ANY AZIMUTH AND AT ANY DISTANCE SHORT OF ANTIPODAL
    C
    C *** AMAX IS THE SEMI-MAJOR AXIS OF THE REFERENCE ELLIPSOID
    C *** FLAT IS THE FLATTENING OF THE REFERENCE ELLIPSOID
    C *** LATITUDES AND LONGITUDES IN RADIANS POSITIVE NORTH AND EAST
    C *** AZIMUTHS IN RADIANS CLOCKWISE FROM NORTH
    C *** GEODESIC DISTANCE S ASSUMED IN UNITS OF SEMI-MAJOR AXIS AMAX
    C
    C *** PROGRAMMED FOR CDC-6600 BY LCDR L.PFEIFER NGS ROCKVILLE MD 20FEB75
    C *** MODIFIED FOR SYSTEM 360 BY JOHN G GERGEN NGS ROCKVILLE MD 750608
    C
    C     IMPLICIT REAL*8 (AMAX-H,O-Z)
    C     COMMON/CONST/PI,RAD
    C     COMMON/ELIPSOID/A,FLAT

    Inputs:
      lat1_rad (float): latitude of station one.
      lon1_rad (float): longitude of station one.
      fAz_rad (float): azimuth from station 1 to station 2.
      s_m (float): sea level distance from station 1 to 2.
    Outputs:
      (float, float, float): lat, lon and back azimuth from station 2 to 1.
    '''

    eps = 0.5e-13
    r = 1 - flat
    tu = r*sin(lat1_rad)/cos(lat1_rad)
    sf = sin(fAz_rad)
    cf = cos(fAz_rad)
    bAz_rad = 0
    if cf != 0:
        bAz_rad = arctan2(tu, cf)*2
    cu = 1/sqrt(tu*tu + 1)
    su = tu*cu
    sa = cu*sf
    c2a = -sa*sa + 1
    x = sqrt((1/r/r - 1)*c2a + 1) + 1
    x = (x - 2)/x
    c = 1 - x
    c = (x*x/4 + 1)/c
    d = (0.375*x*x - 1)*x
    tu = s_m/r/aMax_m/c
    y = tu
    while True:
        sy = sin(y)
        cy = cos(y)
        cz = cos(bAz_rad + y)
        e = cz*cz*2 - 1
        c = y
        x = e*cy
        y = e + e - 1
        y = (((sy*sy*4 - 3)*y*cz*d/6 + x)*d/4 - cz)*sy*d + tu
        if abs(y - c) <= eps:
            break
    bAz_rad = cu*cy*cf - su*sy
    c = r*sqrt(sa*sa + bAz_rad*bAz_rad)
    d = su*cy + cu*sy*cf
    lat2_rad = arctan2(d, c)
    c = cu*cy - su*sy*cf
    x = arctan2(sy*sf, c)
    c = ((-3*c2a + 4)*flat + 4)*c2a*flat/16
    d = ((e*cy*c + cz)*sy*c + y)*sa
    lon2_rad = lon1_rad + x - (1 - c)*d*flat
    bAz_rad = arctan2(sa, bAz_rad) + pi
    return lat2_rad, lon2_rad, bAz_rad

def inverse(p1, e1, p2, e2): # Easier name.
    return gpnhri(p1, e1, p2, e2)

#b::gpnhri
#C
    # SUBROUTINE GPNHRI (A,F,ESQ,PI,P1,E1,P2,E2,AZ1,AZ2,S)    
def gpnhri(p1, e1, p2, e2):

    '''
    Geographic PositioN, Helmert-Rainsford Inverse.
    C
    ********1*********2*********3*********4*********5*********6*********7*
    C
    C NAME:       GPNHRI
    C VERSION:    200208.09
    C WRITTEN BY: ROBERT (SID) SAFFORD
    C PURPOSE:    SUBROUTINE TO COMPUTE HELMERT RAINSFORD INVERSE PROBLEM 
    C 
    C    SOLUTION OF THE GEODETIC INVERSE PROBLEM AFTER T. VINCENTY
    C    MODIFIED RAINSFORD'S METHOD WITH HELMERT'S ELLIPTICAL TERMS
    C    EFFECTIVE IN ANY AZIMUTH AND AT ANY DISTANCE SHORT OF ANTIPOCAL
    C    FROM/TO STATIONS MUST NOT BE THE GEOGRAPHIC POLE.
    C    PARAMETER A IS THE SEMI-MAJOR AXIS OF THE REFERENCE ELLIPSOID
    C    FINV=1/F IS THE INVERSE FLATTENING OF THE REFERENCE ELLIPSOID
    C    LATITUDES AND LONGITUDES IN RADIANS POSITIVE NORTH AND WEST
    C    FORWARD AND BACK AZIMUTHS RETURNED IN RADIANS CLOCKWISE FROM SOUTH
    C    GEODESIC DISTANCE S RETURNED IN UNITS OF SEMI-MAJOR AXIS A
    C    PROGRAMMED FOR IBM 360-195   09/23/75
    C
    C    NOTE - NOTE - NOTE -
    C    1. DO NOT USE FOR MERIDIONAL ARCS AND BE CAREFUL ON THE EQUATOR.
    C    2. AZIMUTHS ARE FROM NORTH(+) CLOCKWISE AND 
    C    3. LONGITUDES ARE POSITIVE EAST(+) 
    C
    C INPUT PARAMETERS:
    C -----------------
    C A       SEMI-MAJOR AXIS OF REFERENCE ELLIPSOID   METERS
    C F       FLATTENING (0.0033528...)
    C ESQ     ECCENTRICITY SQUARED 
    C PI   3.14159...
    C P1      LAT STATION 1                            RADIANS
    C E1      LON STATION 1                            RADIANS
    C P2      LAT STATION 2                            RADIANS
    C E2      LON STATION 2                            RADIANS
    C
    C OUTPUT PARAMETERS:
    C ------------------
    C AZ1     AZI AT STA 1 -> STA 2                    RADIANS
    C AZ2     AZI AT STA 2 -> STA 1                    RADIANS
    C S       GEODETIC DIST BETWEEN STA(S) 1 & 2       METERS
    C
    C LOCAL VARIABLES AND CONSTANTS:
    C ------------------------------
    C AA      CONSTANT FROM SUBROUTINE GPNLOA                  
    C ALIMIT  EQUATORIAL ARC DISTANCE ALONG THE EQUATOR   (RADIANS)
    C ARC     MERIDIONAL ARC DISTANCE LATITUDE P1 TO P2 (IN METERS)     
    C AZ1     AZIMUTH FORWARD                            (IN RADIANS)
    C AZ2     AZIMUTH BACK                               (IN RADIANS)
    C BB      CONSTANT FROM SUBROUTINE GPNLOA                  
    C DLON    TEMPORARY VALUE FOR DIFFERENCE IN LONGITUDE   (RADIANS)   
    C EQU     EQUATORIAL DISTANCE                       (IN METERS)
    C R1,R2   TEMPORARY VARIABLES 
    C S       ELLIPSOID DISTANCE                        (IN METERS)
    C SMS     EQUATORIAL - GEODESIC DISTANCE (S - S) "SMS"       
    C SS      TEMPORARY VARIABLE    
    C TOL0    TOLERANCE FOR CHECKING COMPUTATION VALUE        
    C TOL1    TOLERANCE FOR CHECKING A REAL ZERO VALUE        
    C TOL2    TOLERANCE FOR CLOSE TO ZERO VALUE  
    C TWOPI   TWO TIMES CONSTANT PI               
    C
    C GLOBAL VARIABLES AND CONSTANTS:
    C -------------------------------
    C
    C   MODULE CALLED BY:   GENERAL 
    C
    C   THIS MODULE CALLS:   GPNARC, GPNLOA
    C      LLIBFORE/ DSIN,   DCOS,   DSQRT,  DABS,  DARCTAN2, WRITE
    C
    C   INCLUDE FILES USED:
    C   COMMON BLOCKS USED:  
    C
    C   REFERENCES: MICROSOFT FORTRAN 4.10 OPTIMIZING COMPILER, 1988
    C               MS-DOS OPERATING SYSTEM
    C   COMMENTS:
    ********1*********2*********3*********4*********5*********6*********7*
    ::MODIFICATION HISTORY
    ::197507.05, RWS, VER 00 TENCOL RELEASED FOR FIELD USE
    ::198311.20, RWS, VER 01 MTEN   RELEASED TO FIELD
    ::198411.26, RWS, VER 07 MTEN2  RELEASED TO FIELD
    ::198506.10, RWS, WRK   ENHANCEMENTS RELEASED TO FIELD
    ::198507.22, RWS, CODE   MODIFIED FOR MTEN3
    ::198509.01, RWS, VER 11 MTEN3  RELEASED TO FIELD
    ::198708.10, RWS, CODE   MODIFIED TO USE NEW MTEN4 GPN RECORD FORMAT
    ::199112.31, RWS, VER 20 MTEN4 RELEASED TO FIELD
    ::200001.13, RWS, VER 21 MTEN4 RELEASED TO FIELD
    ::200005.26, RWS, CODE   RESTRUCTURED & DOCUMENTATION ADDED          
    ::200012.31, RWS, VER 23 MTEN5 RELEASED                              
    ::200104.09, RWS, CODE   ADDED TO CALBLIN PROGRAM                      
    ::200208.09, RWS, CODE   ADDED SUBROUTINES GPNARC & GPNLOA            
    ********1*********2*********3*********4*********5*********6*********7*
    E::GPNHRI
    C  -------------------------------
    C    M T E N  (VERSION 3)
    C             (VERSION 4.22)
    C             (VERSION 5.23)
    C  -------------------------------
    C
    C
    '''

    az1_rad = 0
    az2_rad = 0
    s_m = 0
    tol0 = 5e-15
    tol1 = 5e-14
    tol2 = 7e-03
#c
    twopi = 2*pi
#c
#c test the longitude difference with tol1
#c tol1 is approximately 0.000000001 arc seconds
#c
    ss = e2 - e1
    if abs(ss) < tol1:
        #e2 = e2 + tol1  # XXX Unused, commented out. -Mike Markowski
#c           
        s_m = abs(gpnarc(p1, p2))
#c
        if p2 > p1:
            az1_rad, az2_rad = 0, pi
        else:
            az1_rad, az2_rad = pi, 0
        return az1_rad, az2_rad, s_m, 0, 0, 0
#c
#c test for longitude over 180 degrees
#c
    dlon = e2 - e1
#c
    if dlon >= 0:
        if pi <= dlon < twopi:
            dlon -= twopi
        else:
            ss = abs(dlon)
        if pi <= ss < twopi:
            dlon += twopi
#c
    ss = abs(dlon)
    if ss > pi:
#:: write(*,*) '  '
#:: write(*,*) ' longitude difference over 180 degrees  '  
#:: write(*,*) ' turn it around '
        ss = twopi - ss
#c
#c compute the limit in longitude (alimit), it is equal 
#c to twice the distance from the equator to the pole,
#c as measured along the equator (east/ewst)
#c
    alimit = pi*(1 - flat)
#c
#c test for anti-nodal difference     
#c
    if ss >= alimit:
        r1 = abs(p1)
        r2 = abs(p2)
#c
#c   latitudes r1 & r2 are not near the equator
#c
        if (not (r1 > tol2 and r2 > tol2)
            #c
            #c    longitude difference is greater than lift-off point
            #c    now check to see if  "both"  r1 & r2 are on equator
            #c
            and not (r1 < tol1 and r2 > tol2)
            and not (r2 < tol1 and r1 > tol2)):
#c
#c   check for either r1 or r2 just off the equator but < tol2
#c
            if r1 > tol1 or r2 > tol1:
                az1_rad = az2_rad = s_m = 0
                return az1_rad, az2_rad, s_m, 0, 0, 0

#c
#c   compute the azimuth to anti-nodal point
#c
#:: write(*,*) '  '
#:: write(*,*) ' longitude difference beyond lift-off point '  
#:: write(*,*) '  '
#c
            _,az2_rad,_,ao,bo,sms_m = gpnloa(dlon, az1_rad)
#c
#c   compute the equatorial distance & geodetic
#c
            equ_m = aMax_m*abs(dlon) # sets sms_m.
            s_m = equ_m - sms_m
            return az1_rad, az2_rad, s_m, ao, bo, sms_m
#c
#c
    f0 = (1 - flat)
    b = aMax_m*f0
    epsq = eSq/(1 - eSq)
    f2 = flat*flat
    f3 = flat*f2
    f4 = flat*f3
#c
#c the longitude difference 
#c
    dlon = e2 - e1
    ab = dlon
#c
#c the reduced latitudes    
#c
    u1 = f0*sin(p1)/cos(p1)
    u2 = f0*sin(p2)/cos(p2)
#c
    u1 = arctan(u1)
    u2 = arctan(u2)
#c
    su1 = sin(u1)
    cu1 = cos(u1)
#c
    su2 = sin(u2)
    cu2 = cos(u2)
#c
#c counter for the iteration operation
#c
    clon = 0
    q2 = 0
    q4 = 0
    q6 = 0
    r2 = 0
    r3 = 0
    sig = 0
    sinalf = 0
    slon = 0
    ssig = 0
    w = 0
    for kount in range(8):
#c
        clon = cos(ab)
        slon = sin(ab)
#c
        csig = su1*su2 + cu1*cu2*clon
        ssig = sqrt((slon*cu2)*(slon*cu2)
                + (su2*cu1 - su1*cu2*clon)
                * (su2*cu1 - su1*cu2*clon))
#c
        sig = arctan2(ssig, csig)
        sinalf = cu1*cu2*slon/ssig
#c
        w = (1 - sinalf*sinalf)
        t4 = w*w
        t6 = w*t4
#c
#c the coefficients of type a     
#c
        ao = (flat - f2*(1 + flat + f2)*w/4
            + 3*f3*(1 + 9*flat/4)*t4/16 - 25*f4*t6/128)
        a2 = (f2*(1 + flat + f2)*w/4
            - f3*(1 + 9*flat/4)*t4/4 + 75*f4*t6/256)
        a4 = f3*(1 + 9*flat/4)*t4/32 - 15*f4 * t6/256
        a6 = 5*f4*t6/768
#c
#c the multiple angle functions 
#c
        qo = 0
        if w > tol0:
            qo = -2*su1*su2/w
#c
        q2 = csig + qo
        q4 = 2*q2*q2 - 1
        q6 = q2*(4*q2*q2 - 3)
        r2 = 2*ssig*csig
        r3 = ssig*(3 - 4*ssig*ssig)
#c
#c the longitude difference 
#c
        s_m = sinalf*(ao*sig + a2*ssig*q2 + a4*r2*q4 + a6
                * r3*q6)
        xz = dlon + s_m
#c
        xy = abs(xz - ab)
        ab = dlon + s_m
#c
        if xy < 0.5e-13:
            break
#c
#c
#c the coefficients of type b     
#c
    z = epsq*w
#c
    bo = 1 + z*(1/4 + z*(-3/64 + z*(5/256 - z*175/16384)))
    b2 = z*(-1/4 + z*(1/16 + z*(-15/512 + z*35/2048)))
    b4 = z*z*(-1/128 + z*(3/512 - z*35/8192))
    b6 = z*z*z*(-1/1536 + z*5/6144)
#c
#c the distance in meters   
#c
    s_m = b*(bo*sig + b2*ssig*q2 + b4*r2*q4 + b6*r3*q6)
#c
#c first compute the az1 & az2 for along the equator
#c
    if dlon > pi:
        dlon = (dlon - 2*pi)
#c
    if abs(dlon):
        dlon = (dlon + 2*pi)
#c
    az1_rad = pi/2
    if dlon < 0:
        az1_rad = 3*az1_rad
#c
    az2_rad = az1_rad + pi
    if az2_rad > 2*pi:
        az2_rad = az2_rad - 2*pi
#c
#c now compute the az1 & az2 for latitudes not on the equator
#c
    if not (abs(su1) < tol0 and abs(su2) < tol0):
        tana1 = slon*cu2/(su2*cu1 - clon*su1*cu2)
        tana2 = slon*cu1/(su1*cu2 - clon*su2*cu1)
        sina1 = sinalf/cu1
        sina2 = -sinalf/cu2
#c
#c   azimuths from north,longitudes positive east  
#c
        az1_rad = arctan2(sina1, sina1/tana1)
        az2_rad = pi - arctan2(sina2, sina2/tana2)
#c
    if az1_rad < 0:
        az1_rad = az1_rad + 2*pi
#c
    if az2_rad < 0:
        az2_rad = az2_rad + 2*pi
#c
    return az1_rad, az2_rad, s_m, 0, 0, 0

#CB::GPNLOA
#C
def gpnloa(dl, az1):

    '''
    Geographic PositioN, Lift Off Azimuth
    C
    C********1*********2*********3*********4*********5*********6*********7*
    C
    C NAME:     GPNLOA
    C VERSION:   200005.26
    C WRITTEN BY:  ROBERT (Sid) SAFFORD
    C PURPOSE:   SUBROUTINE TO COMPUTE THE LIFF-OFF-AZIMUTH CONSTANTS
    C
    C INPUT PARAMETERS:
    C -----------------
    C AMAX       SEMI-MAJOR AXIS OF REFERENCE ELLIPSOID
    C FLAT       FLATTENING (0.0033528 ... )
    C ESQ         ECCENTRICITY SQUARED FOR REFERENCE ELLIPSOID
    C PI           3.14159...
    C DL           LON DIFFERENCE
    C AZ1         AZI AT STA 1 -> STA 2
    C
    C OUTPUT PARAMETERS:
    C ------------------
    C AZ2         AZ2 AT STA 2 -> STA 1
    C AO           CONST
    C BO           CONST
    C SMS         DISTANCE ... EQUATORIAL - GEODESIC  (S - s)   "SMS"
    C
    C LOCAL VARIABLES AND CONSTANTS:
    C ------------------------------
    C GLOBAL VARIABLES AND CONSTANTS:
    C -------------------------------
    C
    C   MODULE CALLED BY:   GENERAL 
    C
    C   THIS MODULE CALLS:   
    C      LLIBFORE/ DSIN,   cos,   abs,   arcsin 
    C
    C   INCLUDE FILES USED:
    C   COMMON BLOCKS USED:  
    C
    C   REFERENCES: Microsoft FORTRAN 4.10 Optimizing Compiler, 1988
    C               MS-DOS Operating System
    C   COMMENTS:
    C********1*********2*********3*********4*********5*********6*********7*
    C::MODIFICATION HISTORY
    C::1985xx.xx, RWS, CODE   CREATED              
    C::198506.10, RWS, WRK  ENHANCEMENTS RELEASED TO FIELD
    C::198509.01, RWS, VER 11 MTEN3  RELEASED TO FIELD
    C::198512.18, RWS, CODE   MODIFIED FOR MTEN3
    C::198708.10, RWS, CODE   MODIFIED TO USE NEW MTEN4 GPN RECORD FORMAT
    C::199112.31, RWS, VER 20 MTEN4 RELEASED TO FIELD
    C::200001.13, RWS, VER 21 MTEN4 RELEASED TO FIELD
    C::200005.26, RWS, CODE   RESTRUCTURED & DOCUMENTATION ADDED             
    C::200012.31, RWS, VER 23 MTEN5 RELEASED                                 
    C********1*********2*********3*********4*********5*********6*********7*
    CE::GPNLOA
    C ---------------------------
    C    M T E N  (VERSION 3)
    C             (VERSION 4.22)
    C             (VERSION 5.23)
    C ---------------------------
    C
    C
    '''

    tt = 5e-13
#c
    dlon = abs(dl)
    cons = (pi - dlon)/(pi*flat)
#c
#c   compute an approximate az
#c
    az = arcsin(cons)
#c
    t1 = 1
    t2 = (-1/4)*flat*(1 + flat + flat*flat)
    t4 = 3/16*flat*flat*(1 + (9/4)*flat)
    t6 = (-25/128)*flat*flat*flat
#c
    s = 0
    ao = 0
    for iter in range(7):
        s = cos(az)
        c2 = s*s
#c
#c   compute new ao
#c
        ao = t1 + t2*c2 + t4*c2*c2 + t6*c2*c2*c2
        cs = cons/ao
        s = arcsin(cs)
        if abs(s - az) < tt:
            break
#c
        az = s
#c
    az1 = s
    if dl < 0:
        az1 = 2*pi - az1
#c
    az2 = 2*pi - az1
#c
#c   equatorial - geodesic  (s - s)   "sms_m"
#c
    eSqp = eSq/(1 - eSq)
    s = cos(az1)
#c
    u2 = eSqp*s*s
    u4 = u2*u2
    u6 = u4*u2
    u8 = u6*u2
#c
    t1 = 1
    t2 = (1/4)*u2
    t4 = (-3/64)*u4
    t6 = (5/256)*u6
    t8 = (-175/16384)*u8
#c
    bo = t1 + t2 + t4 + t6 + t8
    s = sin(az1)
    sms_m = aMax_m*pi*(1 - flat*abs(s)*ao - bo*(1 - flat))

    # -1 vals are place holders, not true outputs.
    return -1, az2, -1, ao, bo, sms_m
#c

#B::GPNARC
#C
def gpnarc(p1, p2):

    '''
    Geographic PositioN, ARC
    C
    ********1*********2*********3*********4*********5*********6*********7*
    C
    C NAME:     GPNARC
    C VERSION:   200005.26
    C WRITTEN BY:  ROBERT (Sid) SAFFORD
    C PURPOSE:   SUBROUTINE TO COMPUTE THE LENGTH OF AMAX MERIDIONAL ARC 
    C             BETWEEN TWO LATITUDES
    C
    C INPUT PARAMETERS:
    C -----------------
    C AMAX       SEMI-MAJOR AXIS OF REFERENCE ELLIPSOID
    C FLAT       FLATTENING (0.0033528 ... )
    C ESQ         ECCENTRICITY SQUARED FOR REFERENCE ELLIPSOID
    C PI           3.14159...
    C P1           LAT STATION 1
    C P2           LAT STATION 2
    C
    C OUTPUT PARAMETERS:
    C ------------------
    C ARC         GEODETIC DISTANCE 
    C
    C LOCAL VARIABLES AND CONSTANTS:
    C ------------------------------
    C GLOBAL VARIABLES AND CONSTANTS:
    C -------------------------------
    C
    C   MODULE CALLED BY:   GENERAL 
    C
    C   THIS MODULE CALLS:   
    C      LLIBFORE/ OPEN,   CLOSE,  READ,   WRITE,  INQUIRE
    C                abs,   DBLE,   FLOAT,  IABS,   CHAR,   ICHAR
    C
    C   INCLUDE FILES USED:
    C   COMMON BLOCKS USED:  
    C
    C   REFERENCES: Microsoft FORTRAN 4.10 Optimizing Compiler, 1988
    C               MS-DOS Operating System
    C   COMMENTS:
    ********1*********2*********3*********4*********5*********6*********7*
    ::MODIFICATION HISTORY
    ::197507.05, RWS, VER 00 TENCOL RELEASED FOR FIELD USE
    ::198311.20, RWS, VER 01 MTEN   RELEASED TO FIELD
    ::198411.26, RWS, VER 07 MTEN2  RELEASED TO FIELD
    ::1985xx.xx, RWS, CODE   CREATED               
    ::198506.10, RWS, WRK   ENHANCEMENTS RELEASED TO FIELD
    ::198509.01, RWS, VER 11 MTEN3  RELEASED TO FIELD
    ::198512.18, RWS, CODE   MODIFIED FOR MTEN3
    ::198708.10, RWS, CODE   MODIFIED TO USE NEW MTEN4 GPN RECORD FORMAT
    ::199112.31, RWS, VER 20 MTEN4 RELEASED TO FIELD
    ::200001.13, RWS, VER 21 MTEN4 RELEASED TO FIELD
    ::200005.26, RWS, CODE   RESTRUCTURED & DOCUMENTATION ADDED          
    ::200012.31, RWS, VER 23 MTEN5 RELEASED                              
    ********1*********2*********3*********4*********5*********6*********7*
    E::GPNARC
    C ---------------------------
    C    M T E N  (VERSION 3)
    C    M T E N  (VERSION 5.23)
    C ---------------------------
    C 
    '''

    # implicit real*8 (aMax_m-h,o-z)
#c
#   boolean flag
#c
    tt = 5e-15
#c
#c   check for aMax_m 90 degree lookup
#c
    flag = False
#c
    s1 = abs(p1)
    s2 = abs(p2)
#c
    if (pi/2 - tt) < s2 and s2 < (pi/2 + tt):
        flag = True
#c
    if s1 > tt:
        flag = False
#c
    da = (p2 - p1)
    s1 = 0
    s2 = 0
#c
#c   compute the length of aMax_m meridional arc between two latitudes
#c
    e2 = eSq
    e4 = e2*e2
    e6 = e4*e2
    e8 = e6*e2
    ex = e8*e2
#c
    t1 = e2*(3/4)
    t2 = e4*(15/64)
    t3 = e6*(35/512)
    t4 = e8*(315/16384)
    t5 = ex*(693/131072)
#c
    a = 1 + t1 + 3*t2 + 10*t3 + 35*t4 + 126*t5
#c
    if not flag:
#c
        b = t1 + 4*t2 + 15*t3 + 56*t4 + 210*t5
        c = t2 + 6*t3 + 28*t4 + 120*t5
        d = t3 + 8*t4 + 45*t5
        e = t4 + 10*t5
        f = t5
#c
        db = sin(p2*2) - sin(p1*2)
        dc = sin(p2*4) - sin(p1*4)
        dd = sin(p2*6) - sin(p1*6)
        de = sin(p2*8) - sin(p1*8)
        df = sin(p2*10) - sin(p1*10)
#c
#c   compute the s2 part of the series expansion
#c
        s2 = -db*b/2 + dc*c/4 - dd*d/6 + de*e/8 - df*f/10
#c
#c   compute the s1 part of the series expansion
#c
    s1 = da*a
#c
#c   compute the arc length
#c
    arc = aMax_m*(1 - eSq)*(s1 + s2)
    return arc
#c

def plh2Xyz_NDS(lat_rad, lon_rad, elev):

    '''
    ********1*********2*********3*********4*********5*********6*********7*********
    name: plh2xyz
    version: 9602.20
    written by: C. Goad
    purpose: converts elliptic lat, lon, hgt to geocentric X, Y, Z
     
    input parameters
    ----------------
    A semi-major axis of ellipsoid [units are of distance_m]
    FL flattening of ellipsoid [unitless]
    plh[] ellipsoidal coordinates of point, in geodetic latitude, longitude east of
    Greenwich, and
    height [units for latitude=plh[0] and longitude=plh[1] are
    in degrees height=plh[2] are distance_m and will be the same as those of
    the input parameters]
     
    output parameters
    -----------------
    xyz[] geocentric Cartesian
    coordinates [units are of distance_m]
     
     
    local variables and constants
    -----------------------------
     
    global variables and constants
    ------------------------------
     
     
    called by:
    ------------------------------
     
    calls:
    ------------------------------
     
    include files:
    ------------------------------
     
    references:
    ------------------------------
    Escobal, "Methods of Orbit Determination", 1965, Wiley & Sons, Inc., pp. 27-29.
     
    comments:
    ------------------------------
    This routine was stripped from one called tlate which converted
    in both directions. A better routine (not iterative) was gotten
    for XYZ to PLH in turn, this was specialized and made to match.

    see also:
    ------------------------------
    xyz2plh

     *******1*********2*********3*********4*********5*********6*********7*********
    :modification history
    :8301.00, CG, Creation
    :9406.16, MSS, Conversion to C.

    Inputs:
      lat_rad latitude in radians.
      lon_rad longitude in radians.
      elev_m height in meters.
    Output:
      np.array([x, y, z]) in meters.
    '''

    sinLat = sin(lat_rad)
    g1 = aMax_m/sqrt(1 - eSq*sinLat*sinLat)
    g2 = g1*(1 - flat)*(1 - flat) + elev_m
    g1 += elev_m

    x = g1*cos(lat_rad)
    y = x*sin(lon_rad)
    x = x*cos(lon_rad)
    z = g2*sinLat

    return np.array([x, y, z])

def plhToXyz(lat_rad, lon_rad, elev_m):

    sinLat = sin(lat_rad)
    r = aMax_m/sqrt(1 - eSq*sinLat*sinLat)
    f1 = (r + elev_m)*cos(lat_rad)
    f2 = (r + elev_m) - eSq*r

    x_m = f1*cos(lon_rad)
    y_m = f1*sin(lon_rad)
    z_m = f2*sinLat

    return np.array([x_m, y_m, z_m])

def xyz2plh(x_m, y_m, z_m):

    '''
    *<pre>
    C********1*********2*********3*********4*********5*********6*********7**
    C Name: xyz2plh Version: 9602.17 Author: B. Archinal (USNO) Purpose:
    C Converts XYZ geocentric coordinates to Phi (latitude), Lambda
    C (longitude), H (height) referred to an ellipsoid of semi-major axis A and
    C flattening FL.
    C 
    C Input: ----------- A semi-major axis of ellipsoid [units are of
    C distance_m] FL flattening of ellipsoid [unitless] xyz[] geocentric
    C Cartesian coordinates [units are of distance_m]
    C 
    C Output: ----------- plh[] ellipsoidal coordinates of point, in geodetic
    C latitude, longitude east of Greenwich, and height [units for
    C latitude=plh[0] and longitude=plh[1] are in degrees height=plh[2] are
    C distance_m and will be the same as those of the input parameters]
    C 
    C Local: ----------- B semi-minor axis of ellipsoid [same units as A]
    C 
    C Global: -----------
    C 
    C Notes: ----------- This routine will fail for points on the Z axis, i.e.
    C if X= Y= 0 (Phi = +/- 90 degrees).
    C 
    C Units of input parameters `A' and `xyz' must be the same.
    C 
    C References: ----------- Borkowski, K. M. (1989). "Accurate algorithms to
    C transform geocentric to geodetic coordinates", *Bulletin Geodesique*, v.
    C 63, pp. 50-56.
    C 
    C Borkowski, K. M. (1987). "Transformation of geocentric to geodetic
    C coordinates without approximations", *Astrophysics and Space Science*, v.
    C 139, n. 1, pp. 1-4. Correction in (1988), v. 146, n. 1, p. 201.
    C 
    C An equivalent formulation is recommended in the IERS Standards (1995),
    C draft.
    C 
    C********1*********2*********3*********4*********5*********6*********7**
    C Modification History: 9007.20, BA, Creation 9507,21, JR, Modified for use
    C with the page programs 9602.17, MSS, Converted to C.
    C********1*********2*********3*********4*********5*********6*********7**
    C </pre>
    C 
    C 
      Some variable and parameter changes:
      
      COMMON var A to  AMAX
      iCOMMON var FL to FLAT.
      
      @param x_m standard terrestrial reference system coordinate of location.
      @param y_m standard terrestrial reference system coordinate of location.
      @param z_m standard terrestrial reference system coordinate of location.
      @return coordinate converted to WGS84 lat/lon/el where units are in
      degrees and meters above sea level.
    '''

    # 1.0 compute semi-minor axis and set sign to that of z in order to get
    # sign of phi correct
    b = aMax_m*(1 - flat)
    if z_m < 0:
        b = -b
    # 2.0 compute intermediate values for latitude
    r = sqrt(x_m*x_m + y_m*y_m)
    e = (b*z_m - (aMax_m*aMax_m - b*b))/(aMax_m*r)
    f = (b*z_m + (aMax_m*aMax_m - b*b))/(aMax_m*r)
    # 3.0 find solution to: t^4 + 2*e*t^3 + 2*f*t - 1 = 0
    p = (4/3)*(e*f + 1)
    q = 2*(e*e - f*f)
    d = p*p*p + q*q

    if d >= 0:
        v = (sqrt(d) - q)**(1/3) - (sqrt(d) + q)**(1/3)
    else:
        v = 2*sqrt(-p)*cos(arccos(q/(p*sqrt(-p)))/3)
    # 4.0 improve v note: not really necessary unless point is near pole
    if v*v < abs(p):
        v = -(v*v*v + 2*q)/(3*p)
    g = (sqrt(e*e + v) + e)/2
    t = sqrt(g*g + (f - v*g)/(2*g - e)) - g

    lat_rad = arctan((aMax_m*(1 - t*t))/(2*b*t))
    # 5.0 compute height above ellipsoid
    elev = (r - aMax_m*t)*cos(lat_rad) + (z_m - b)*sin(lat_rad)
    # 6.0 compute longitude east of Greenwich
    zlong = arctan2(y_m, x_m)
    if zlong < 0:
        zlong = zlong + 2*pi

    lon_rad = zlong
    # 7.0 convert latitude and longitude to degrees
    return np.array([np.degrees(lat_rad), np.degrees(lon_rad), elev])

def xyzToWgs84_borkowski(x, y, z):

    '''
    Info from:
     
    IERS TECHNICAL NOTE 21 IERS Conventions (1992) Dennis D. McCarthy US
    Naval Observatory
    '''
    lambda0 = arctan2(y, x)
    r = sqrt(x*x + y*y)
    v = geod(r, z) # Calculates gedoFi & geodH in next line.
    return np.array([lambda0, v[0], v[1]])

def geod(r, z):

    '''
    Algorithm from:
     
    "Accurate algorithms to transform geocentric to geodetic coordinates,"
    Author: BORKOWSKI, KAZIMIERZ M. Affiliation: AA (Torun Radio Astronomy
    Observatory, Poland) Journal: Bulletin Geodesique, vol. 63, no. 1, 1989,
    p. 50-56.

    See http://www.astro.uni.torun.pl/~kb/Papers/geod/Geod-BG.htm
    and http://www.astro.uni.torun.pl/~kb/Papers/geod/Geod-GK.htm

    Polish version of paper includes Fortran source which I converted
    below making as few changes as possible. - Mike Markowski

    Inputs:
      r (float): m, equatorial component.
      z (float): m, polar component.
    Outputs"
      (np.array[2]): geodFi, geodH, geodetic coordinates of latitude in
        radians and height in meters.
    '''

    b = sign(aMax_m - aMax_m/fRec, z)
    if r == 0:
        return None
    e = ((z + b)*b/aMax_m - aMax_m)/r
    f = ((z - b)*b/aMax_m + aMax_m)/r
    # Find solution to: t**4 + 2*e*t**3 + 2*f*t - 1 = 0
    p = (e*f + 1)*4/3
    q = (e*e - f*f)*2
    d = p*p*p + q*q
    # Improve the accuracy of numeric values of v
    if d >= 0:
        s = sqrt(d) + q
        s = sign(np.exp(np.log(abs(s))/3), s)
        v = p/s - s
        v = -(q + q + v*v*v)/(3*p)
    else:
        v = 2*sqrt(-p)*cos(arccos(q/p/sqrt(-p))/3)
    g = 0.5*(e + sqrt(e*e + v))
    t = sqrt(g*g + (f - v*g)/(g + g - e)) - g

    geodFi = arctan((1 - t*t)*aMax_m/(2*b*t))
    geodH = (r - aMax_m*t)*cos(geodFi) + (z - b)*sin(geodFi)
    return np.array([geodFi, geodH])

def sign(a, b):

    '''
    Implement the Fortran SIGN(A, B) function.
    SIGN(A,B) returns the value of A with the sign of B.
    '''
    return abs(a) if b >= 0 else -abs(a)

def getrad(d, m, sec, isign):
# *** convert deg, min, sec to radians
    return isign * np.radians(d + m/60 + sec/3600)
