#!/usr/bin/env python

'''
This implements a simple noise model estimating noise in a receiver and
environmental noise.  The environmental curves are taken from Table 1 in "An
Updated Noise Model for Use in IONCAP," NTIA Report 87-212, Spaulding, 1987.

Mike Markowski, mike.ab3ap@gmail.com
Apr 2023
'''

import numpy as np
import sys

BOLTZ = 1.380648813e-23 # Joules/Kelvin, Boltzmann's constant

genois = {
	# Fgure 5.  From VAOCAP's genois.for.  Valid for up to 300 MHz.
	# Category      m     b         Valid from 0.3 to 250 MHz.
	'business'   : [0.3, 250, -27.7, 76.8],
	'highway'    : [0.3, 250, -27.7, 73.0],
	'residential': [0.3, 250, -27.7, 72.5],
	'park'       : [0.3, 250, -27.7, 69.3],
	'rural'      : [0.3, 250, -27.7, 67.2],
	'quiet'      : [0.3, 30, -28.6, 53.6], # Valid form 0.3 to 39 MHz.
	'galactic'   : [10, 170, -23.0, 52.0]  # Valid from 10 to about 170 MHz.
}

def noise_dBm(f_MHz, neb_Hz, nf_dB, env, rcvr_K=290):

#	cosmic_K = 2.724 # Cosmic noise temperature.
#	cosmic_dB = BOLTZ*cosmic_K # 1 Hz bandwidth, -194 dBm.
	
	To_K = 290 # GENOIS graphs are relative to kT0, or BOLTZ*To_K.

	#   I n t e r n a l
	nf = 10**(nf_dB/10)  # Convert noise figure dB to noise factor.
	nf_K = To_K*(nf - 1) # Equivalent noise temperature.
	int_W = BOLTZ*(rcvr_K + nf_K)*neb_Hz # Internal noise power.

	#   E x t e r n a l
	f0_MHz, f1_MHz, m, b = genois['galactic']
	if (f0_MHz <= f_MHz <= f1_MHz):
		galactic_dB = m*np.log10(f_MHz) + b # dB above kT.
		galactic_K = To_K * 10**(galactic_dB/10)
	else:
		galactic_K = 0

	f0_MHz, f1_MHz, m, b = genois[env]
	if (f0_MHz <= f_MHz <= f1_MHz):
		local_dB = m*np.log10(f_MHz) + b # dB above kT.
		# XXX Freq adjustment, Eq 12 in Spaulding, should be performed here!
		local_K = To_K * 10**(local_dB/10) # Temp
	else:
		local_K = 0
	ext_W = BOLTZ*(galactic_K + local_K)*neb_Hz

	nTotal_dBm = 10*np.log10(int_W + ext_W) + 30
	return nTotal_dBm

# m a i n
#
# A simple program to test the subroutine.

def main(argv):

	prog = argv[0]
	argv = argv[1:]

	print('For f=146.985 MHz, NEB=15 kHz, NF=4 dB, residential env')
	print('noise_dBm(146.985, 15e3, 4, \'residential\')')
	print('Noise power: %.1f dBm' % noise_dBm(146.985, 15e3, 4, 'residential'))

if __name__ == '__main__':
	main(sys.argv)

