#!/usr/bin/env python

'''
rfl (Radio Frequency Link) uses models for antenna gain, noise power and
path loss to predict link budgets.  Power at receivers can be presented in
units of uV, dBuV, dBm, and S/N either on a coverage map or a csv (comma
separated value) file for all pairwise tx-to-rx links.

Simulations are set up using a configuration file.  A python dictionary is
created from the configuration file, so adding a GUI that creates a
dictionary should be straightforward.  But GUIs are time consuming to
create and my focus is on the RF aspects.

A default configuration can be created with the -d command line option and
contains comments that hopefully are self-explanatory.  The danger is that
what seems obvious to the me is in fact not so obvious.  :-o  Please help
improve this code.  It needs it!

Tnx es 73,
Mike Markowski ab3ap
Apr 2023
'''

import config, earth, itm, noise, nsma, prof

from matplotlib.collections import PatchCollection
from matplotlib.patches import Polygon
from scipy.stats import norm
import guiConnect as gui
import matplotlib.pyplot as plt
import numpy as np
import sys

ants = {} # NSMA antenna data file dictionaries.
progress = None
rcvrs = []
xmtrs = []

mem = {} # Memory for comparisons with previous variable values.
mem['service'] = mem['climate'] = -1

dhItm = 0
dhItmN = 0

def area_dB(d, dist_m):
    '''Use ITM model to predict RF path loss over terrain using area mode.
    This subroutine's purpose is to gather info needed to make a call to
    lrArea() followed by one to aVar().  ITM's area mode does not have
    terrain profile data and characterizes local terrain by a single number,
    the terrain roughness factor.  If reliability and confidence levels are
    provided, a call to aVar() augments median path loss with these levels.

    Inputs:
      d (dictionary): dictionary of the rfl configuration file.
      dist_m (float): distance from transmitter to calculate path loss.
    Output:
      loss_dB (float): path loss as calculated by ITM.
      elTx_rad (float): RF takeoff angle in radians of transmitted RF wave.
      elRx_rad (float): RF takeoff angle in radians of received RF wave.
    '''

    global rcvrs, xmtrs

    itm.dictionaries()  # Set ITM values to defaults.

    if len(xmtrs) == 0:
        xmtrs, rcvrs = stnGroups(d)
    if len(xmtrs) > 1:
        print('Note:         %d receivers found, but only using %s.'
            % (len(rcvrs), rcvrs[0]))
    if len(xmtrs) > 1:
        print('Note:         %d transmitters found, but only using %s.'
            % (len(xmtrs), xmtrs[0]))
    stnRxName = rcvrs[0]
    stnTxName = xmtrs[0]
    stnTx = d['station'][stnTxName]
    stnRx = d['station'][stnRxName]
    itm.prop['hg'][0] = stnTx['hgt'] # hg is Height above Ground.
    itm.prop['hg'][1] = stnRx['hgt']
    itm.prop['dh'] = d['rfl']['dh']  # Terrain roughness factor.
    sitingTx = stnTx['site']         # Care of tx station siting.
    sitingRx = stnRx['site']         # Care of rx station siting.
    service = stnTx['service']       # Transmission mode.
    if d['rfl']['interference'] == 'yes':
        sevice += 20

    # Get environmental values needed by ITM.
    elev_m = (stnTx['hgt'] + stnRx['hgt'])/2 # Avg elev above sea level.
    env    = d['env'][stnTx['env']]
    climate = env['climate'] # Receiver climate.
    epsR    = env['eps']     # Relative ground permittivity, F/m.
    N0      = env['N0']      # Sea level refractivity in given region.
    sigma   = env['sigma']   # Relative ground conductivity, Siemens.
    polAnt  = stnTx['pol']   # Vertical antenna polarization.
    nName   = stnTx['noise']
    f_MHz   = float(d['noise'][nName]['f'])

    # Calculate median path loss.  aVar() follow up calls are recommended.
    itm.lrPrep(f_MHz, elev_m, N0, polAnt, epsR, sigma)
    itm.lrArea([sitingTx, sitingRx], climate, service)
    itm.lrProp(dist_m) # Area mode distance.

    # Retrieve answers from ITM dictionary.
    elTx_rad = itm.prop['the'][0] # RF takeoff angle at transmitter.
    elRx_rad = itm.prop['the'][1] # RF takeoff angle at receiver.

    # Apply confidence and reliability levels.
    if 'conf' in d['rfl'].keys():
        qc = d['rfl']['conf']
        if not 'loc' in d['rfl'].keys():
            print('Warning: fraction of locations not specified, using 0.5.')
            d['rfl']['loc'] = 50
        if not 'time' in d['rfl'].keys():
            print('Warning: fraction of time not specified, using 0.5.')
            d['rfl']['time'] = 50
        qc = d['rfl']['conf'] # Confidence.
        qt = d['rfl']['time'] # Quantile of time (availability).
        ql = d['rfl']['loc']  # Quantile of location.
        zc = norm.isf(qc/100)
        zt = norm.isf(qt/100)
        zl = norm.isf(ql/100)
        med_dB, _ = itm.aVar(zt, zl, zc)
    else:
        med_dB = itm.prop['aref'] # lrProp() result.
    fs_dB = freespace_dB(None, 1e6*f_MHz, dist_m)
    loss_dB = fs_dB + med_dB
    return loss_dB, elTx_rad, elRx_rad

def cli(argv):
    '''
    Usage: rfl [-c file] [-d file]

       -c file: read configuration from 'file'.
       -d file: write default configuration to 'file'.
    '''

    prog = argv[0]
    argv = argv[1:]

    d = None
    i = 0
    while i < len(argv):
        arg = argv[i]
        match arg:
            case '-c': # Configuration file.
                i += 1
                d = config.cfgParse(config.cfgRead(argv[i]))
            case '-d': # Default configuration file creation.
                i += 1
                config.cfgWrite(argv[i])
                sys.exit(0)
            case _: # Error
                usage()
                sys.exit(0)
        i += 1
    return d

def contour(d, snr_dB, region_rad, pAng, pGain, gui=False):
    '''Given results of an rfl simulation, plot the results as isoline
    (contour) levels.

    Inputs:
      d (dictionary): dictionary of the rfl configuration file.
      snr_dB (float[][]): 2D grid of link budget values.
      region_rad (float[4]): [lonW,lonE,latS,latN] defining map region.
      pAng (float[]): antenna pattern angles in radians.
      pGain (float[]): antenna pattern gains.
      titlePlot (string): title of coverage map plot.
      titleCbar (string): title of coverage map color bar.
    Output:
      none
    '''

    titlePlot = d['coverage']['title']
    subtitle = ''
    if True:
        allTx, allRx = stnGroups(d)
        if len(allTx) == 1:
            if titlePlot != '':
                titlePlot += '\n'
            qc = d['rfl']['conf'] # Confidence.
            qt = d['rfl']['time'] # Quantile of time (availability).
            ql = d['rfl']['loc']  # Quantile of location.
            stnTx = d['station'][allTx[0]]
            match stnTx['service']: # Transmission mode.
                case 0:
                    subtitle = 'First Try Probability\n'
                    subtitle += '%1.f%% Confidence' % qc
                case 1:
                    subtitle = 'Random Broadcast Receiver\n'
                    subtitle += '%1.f%% Confidence,  ' % qc
                    subtitle += '%1.f%% of Time Availability' % qt 
                case 2:
                    subtitle = 'Mobile Service\n'
                    subtitle += '%1.f%% Confidence,  ' % qc
                    subtitle += '%1.f%% Reliability' % qt
                case 3:
                    subtitle = 'Broadcast Service\n'
                    subtitle += '%1.f%% Confidence,  ' % qc
                    subtitle += 'Reliability %1.f%% of Time,  ' % qt
                    subtitle += '%1.f%% of Locations' % qt
            titlePlot += subtitle

    titleCbar = units = d['rfl']['units']
    if units == 'SNR':
        titleCbar = 'S/N (dB)'

    lonW_rad, lonE_rad, latS_rad, latN_rad = region_rad
    lonW = np.degrees(lonW_rad)
    lonE = np.degrees(lonE_rad)
    latS = np.degrees(latS_rad)
    latN = np.degrees(latN_rad)

    if gui:
        # ion() needed to avoid warning with plt.show():
        #    QCoreApplication::exec: The event loop is already running
        plt.ion()
    fig = plt.figure(figsize=(12,9))
    ax = fig.add_subplot(1, 1, 1)

    lats = np.linspace(latS, latN, snr_dB.shape[0])
    lons = np.linspace(lonW, lonE, snr_dB.shape[1])
    X, Y = np.meshgrid(lons, lats)

    if 'coverage' in d.keys() and 'levels' in d['coverage'].keys():
        levels = d['coverage']['levels']
    else:
        levels = None

    if d['coverage']['contour'] == 'fill':
        cp = plt.contourf(X, Y, snr_dB, levels=levels, cmap='Spectral_r')
    else:
        cp = plt.contour(X, Y, snr_dB, levels=levels, cmap='Spectral_r')
    plt.colorbar(cp, label=titleCbar)

    for stn in d['station'].keys():
        # Plot each station.
        if 'lon' in d['station'][stn].keys() and d['station'][stn]['lon'] != '':
            lon_deg = np.degrees(d['station'][stn]['lon'])
            lat_deg = np.degrees(d['station'][stn]['lat'])
            height = abs(latN - latS)
            width = abs(lonE - lonW)
            side = 0.005*min(height, width)

            # Plot transmitter antenna pattern.
            if ('tx' in d['station'][stn].keys()
                and not pAng is None
                and d['coverage']['pattern'] == 'yes'):
                frac = 0.08
                height = abs(frac*(latN - latS))
                width = abs(frac*(lonE - lonW))

                # Scale and convert polar pattern to x,y.
                pGain /= np.max(pGain)
                x = width*pGain*np.cos(pAng)
                y = height*pGain*np.sin(pAng)
                pts = np.vstack((x,y)).T

                # Rotate and translate pattern,
                # XXX Unsure how it's already rotated ccw off east azAnt_rad.
                azAnt_rad = np.pi/2 - 2*d['station'][stn]['az']
                c = np.cos(azAnt_rad)
                s = np.sin(azAnt_rad)
                M = np.array([
                    [c, -s],
                    [s,  c]
                ])
                pts = np.array(
                    [np.matmul(M, pts[i]) for i in range(x.size)])

                pts[:,0] += lon_deg # Translation to lat/lon.
                pts[:,1] += lat_deg
                polygon = Polygon(pts)

                # Draw it!
                patches = []
                patches.append(polygon)
                p = PatchCollection(patches, alpha=0.2, color='green')
                ax.add_collection(p)
                ax.plot(pts[:,0], pts[:,1], lw=0.5, color='green')

            # Create, draw and label station.
            circle1 = plt.Circle((lon_deg, lat_deg), side, color='lightcyan')
            circle2 = plt.Circle((lon_deg, lat_deg), side, color='black',
                fill=False, lw=0.5)
            plt.text(side+lon_deg, side+lat_deg, stn)
            # Draw it!
            patches = []
            ax.add_patch(circle1)
            ax.add_patch(circle2)

#   fig.suptitle(titlePlot, fontsize=12)
#   ax.set_title(subtitle, fontsize=10)
    ax.set_title(titlePlot)
    ax.set_xlabel('Longitude')
    ax.set_ylabel('Latitude')
    pngFile = d['rfl']['outfile'].strip()
    pngFile = 'rfl.png' if pngFile=='' else pngFile
    plt.savefig(pngFile)
    if not gui:
        plt.show()

def cut(adf, f_MHz, pol, azAnt_rad, elAnt_rad):
    ''' Taking into account an antenna's azimuth and elevation, create a cut
    at 0 degrees elevation.  The cut can be plotted to aid visulation.

    Inputs:
      adf (dictionary): dictionary of NSMA antenna data file information.
      f_MHz (float): carrier frequency.
      pol (string): antenna polarity pair.  E.g., 'V/V', 'H/V', etc., meaning
        polarized port response to polarized signal.
      azAnt_rad (float): azimuth of antenna, clockwise from true north.
      elAnt_rad (float): elevation of antenna.
    Output:
      angles_rad (float): angles in interpolated cut at 0 deg elevation.
      gains (float): gains in interpolated cut at 0 deg elevation.
      units (string): dBi, dBd, dBr, or lin.
    '''

    gains = []
    angles_rad = np.linspace(0, 2*np.pi, 60, endpoint=False)
    for angle_rad in angles_rad:
        g, units = nsma.gain(adf, f_MHz, pol,
            azAnt_rad, elAnt_rad, angle_rad, 0)
        gains.append(g)
    gains = np.array(gains)
    gains -= np.min(gains) # Must have no negative gains for plotting.
    return angles_rad, gains, units

def dBm_to_uV(P_dBm, Z_ohms=50): # Convert dBm to uV.
    # P=IV, so V = P/I = P/(V/Z) = PZ/V.  So V**2 = PZ.
    # If Z=50, V**2 = 50*P_dBm = 50 * 10**(dBm/10) - 30)
    # V = sqrt(50 * 10**((dBm-30)/10))
    return np.sqrt(Z_ohms * 10**((P_dBm-30)/10)) * 1e6

def uV_to_dBm(rms_uV, Z_ohms=50): # Convert uV to dBm.
    # P = IV = V V/Z
    # If Z=50, P = (uV/1e6)**2/50
    # dBm = 10*log10(P) + 30
    return 10*np.log10((rms_uV/1e6)**2/Z_ohms) + 30

def stnGroups(d):
    '''From an rfl dictionary return lists of transmitting and receiving
    stations.

    Input:
      d (dictionary): dictionary of rfl configuration.
    Output:
      tx, rx (lists): lists of names of transmitting and receiving stations.
    '''

    rx = []
    tx = []
    for stn in d['station'].keys():
        if 'tx' in d['station'][stn].keys():
            tx.append(stn)
        else:
            rx.append(stn)
    return tx, rx # Stations grouped as transmitters, receivers.

def freespace_dB(L_dB, f_Hz, r_m):
    '''Solve a freespace path loss problem.  Set the parameter to None that
    is to be solved for.

    Inputs:
      L_dB (float): free space path loss in dB.
      f_MHz (float): frequency in MHz.
      r_m (float): distance in meters to calculate path loss.
    Output:
      (float): freespace path loss in dB.
    '''

    c = 299792458 # m/s, speed of light in vacuum, Wikipedia.
    if L_dB == None: # Solve for loss.
        return 20*np.log10(4*np.pi*r_m*f_Hz/c)
    elif f_MHz == None: # Solve for frequency.
        return 10**(L_dB/20) * c/(4*np.pi*r_m)
    else: # Solve for distance.
        return 10**(L_dB/20) * c/(4*np.pi*f_Hz)

def gainAnt(d, stn, azRf_rad=None, elRf_rad=None, vswr=None):
    '''If the rfl configuration file does not mention the 'ant' (antenna)
    keyword, an isotropic antenna is assumed.  If the 'ant' variable has a
    value that is a number, then an isotropic of that gain is used.

    The typical case uses an antenna pattern in NSMA format.  The antenna
    pattern boresight runs along the x-axis and is is pointed at some
    azimuth and elevation.  The RF ray has its own az/el relative to
    the oriented pattern.  Rather than rotate all points in the pattern,
    it is simpler to reverse rotate the RF ray.

    Inputs:
      d (dictionary): dictionary of rfl configuration.
      stn (dictionary): dictionary of station description.
      azRf_rad (float): optional, azimuth of RF ray piercing pattern.
      elRf_rad (float): optional, elevation of RF ray piercing pattern.
      vswr (float): optional, vswr of antenna system.
    Output:
      g_dB (float): gain of antenna for RF ray at specified az/el.
      pAng (float[]): 0 deg elevation cut angles, used for plotting.
      pGain (float[]): 0 deg elevation cut gains, used for plotting.
    '''

    g_dB = [0,0]
    pol = ['v', 'v']

    for i in [0,1]:
        if not 'ant' in stn[i].keys(): # Isotropic, 0 dBi.
            g_dB[i] = 0 # dBi.
        elif isNumber(stn[i]['ant']): # Isotropic, dBi.
            g_dB[i] = float(stn[i]['ant'])
        else: # Use NSMA antenna data file.
            pol[i] = stn[i]['pol']
    pol = pol[0] + '/' + pol[1]
    nName = stn[1]['noise']
    nRx = d['noise'][nName]
    f_MHz  = nRx['f'] # String version of freq.

    pAng = pGain = None
    for i in [0,1]:
        if not 'ant' in stn[i].keys() or isNumber(stn[i]['ant']):
            continue
        adfFile = d['rfl']['ant'] + '/' + stn[i]['ant']
        if adfFile in ants.keys():
            adf = ants[adfFile]
        else:
            ants[adfFile] = adf = nsma.nsma(adfFile)
        if 'az' in stn[i].keys():
            azAnt_rad = stn[i]['az']
            elAnt_rad = stn[i]['el']
        else:
            azAnt_rad = stn[i]['az'] = 0
            elAnt_rad = stn[i]['el'] = 0
        g_dB[i], units = nsma.gain(adf, f_MHz, pol,
            azAnt_rad, elAnt_rad, azRf_rad[i], elRf_rad[i])
        g_dB[i] = g_dB[i]+2.15 if units=='dBd' else g_dB[i] # dBd to dBi.

        if 'tx' in stn[i].keys():
            # Get antenna pattern for plot.
            pAng, pGain, _ = cut(adf, f_MHz, pol, azAnt_rad, elAnt_rad)

    # VSWR mismatch loss.
    if vswr != None:
        vswrLoss_dB = -10*np.log10(1 - ((vswr-1)/(vswr+1))**2)
        g_dB -= vswrLoss_dB
    return g_dB, pAng, pGain # dBi gain and horizontal pattern.

def isNumber(s):
    '''True/False if string s is a float.
    '''
    try:
        float(s)
        return True
    except ValueError:
        return False

def linkBudget(d, stnTxName, stnRxName):
    '''Perhaps the most important subroutine in this program, it performs a
    link budget between a tx and rx station.

    Inputs:
      d (dictionary): dictionary of rfl configuration.
      stnTxName (string): name of transmit station.
      stnRxName (string): name of receive station.
    Outputs:
      rx_dBm (float): signal power at receiver in dBm, the link budget result.
      gain_dB (float[2]): gains in dBi of tx and rx station.
      pl_dB (float): ITM path loss between two stations.
      patAng (float[]): 0 deg elevation cut angles, used for plotting.
      patGain (float[]): 0 deg elevation cut gains, used for plotting.
      domMode (string): name of dominant mode of propagation: single horizon,
        double horizon, diffraction, tropospheric scatter.
    '''

    # Retrieve station values needed for link budget.
    stnRx = d['station'][stnRxName]
    stnTx = d['station'][stnTxName]
    latTx_rad = stnTx['lat']
    lonTx_rad = stnTx['lon']
    sysRx_dB  = stnRx['sys'] if 'sys' in stnRx.keys() else 0
    sysTx_dB  = stnTx['sys'] if 'sys' in stnTx.keys() else 0
    sys_dB    = sysRx_dB + sysTx_dB
    tx_dBm    = stnTx['tx']
    if d['rfl']['output'] == 'coverage':
        latRx_rad = stnRx['latCell']
        lonRx_rad = stnRx['lonCell']
    else:
        latRx_rad = stnRx['lat']
        lonRx_rad = stnRx['lon']

    # Retrieve receiver noise values needed for link budget.
    nRxName = stnRx['noise']
    nRx = d['noise'][nRxName]
    f_MHz  = nRx['f']   # String, frequency of signal.

    #  Rfl data.
    terrain = d['rfl']['terrain']
    dted = d['rfl']['dted']

    # Get:
    if d['rfl']['mode'] == 'profile':
        # 1. terrain profile.
        rng, pfl = prof.pfl(
            [latTx_rad,lonTx_rad], [latRx_rad,lonRx_rad], terrain, dted)
        # 2. path loss and RF takeoff angles.
        pl_dB, elTx_rad, elRx_rad = pathloss_dB(
            rng, pfl, d, stnTxName, stnRxName)
        res1, res2 = itm.propMode(pfl)

        match res2:
            case 'd':
                pmode = 'diffraction ('
            case 't':
                pmode = 'tropo scatter ('
            case '':
                pmode = ''
        match res1:
            case 'l':
                pmode += 'los'
            case 's':
                pmode += 'single horizon)'
            case 'd':
                pmode += 'double horizon)'

    else: # 'area' mode.
        # 1. distance between stations.
        _,_,dist_m,_,_,_ = earth.inverse(
            latTx_rad, lonTx_rad, latRx_rad, lonRx_rad)
        # 2. path loss and RF takeoff angles.
        pl_dB, elTx_rad, elRx_rad = area_dB(d, dist_m)
        pmode = 'n/a'
    # 3. antenna gains.
    # Tx to rx angle off true north.
    azTx_rad,azRx_rad,s,_,_,_ = earth.inverse(
        latTx_rad, lonTx_rad, latRx_rad, lonRx_rad)
    gain_dB, patAng, patGain = gainAnt(
        d, [stnTx,stnRx], [azTx_rad,azRx_rad], [elTx_rad,elRx_rad])

    # Calculate: link budget.
    rx_dBm = tx_dBm + gain_dB[0] + gain_dB[1] - pl_dB - sys_dB
    return rx_dBm, gain_dB, pl_dB, patAng, patGain, pmode

def coverage(d):
    '''Loop through a grid of cells on a map and perform a link budget
    between transmitter and each cell.

    Input:
      d (dictionary): dictionary of rfl configuration.
    Output:
      rxS_dBm (float[][]): signal power in each grid cell.
      rxN_dBm (float[][]): noise piwer in each grid cell.
      area_rad (float[4]): [lonW,lonE,latS,latN] of grid.
      pAng (float[]): 0 deg elevation cut angles, used for plotting.
      pGain (float[]): 0 deg elevation cut gains, used for plotting.
    '''

    global progress

    allTx, allRx = stnGroups(d)
    if len(allRx) > 1:
        print('Note:         %d receivers found, '
            'using rx & env data only from %s.' % (len(allRx), allRx[0]))
    if len(allTx) > 1:
        print('Note:         %d transmitters found, using only %s.'
            % (len(allTx), allTx[0]))
    if len(allTx) == 0 or len(allRx) == 0:
        if len(allTx) == 0:
            print('No transmitters found.')
        if len(allRx) == 0:
            print('No receivers found.')
        sys.exit(1)
        return None, None, None, None, None

    # Retrieve station values needed for link budget.
    stnRxName = allRx[0]
    stnTxName = allTx[0]
    stnRx = d['station'][stnRxName]
    stnTx = d['station'][stnTxName]

    # Find bounding lat/lons.
    if 'lat' in d['coverage'].keys() and d['coverage']['lat'] != '':
        latCtr_rad = d['coverage']['lat']
        lonCtr_rad = d['coverage']['lon']
    else:
        lat0_rad = lon0_rad = np.Inf
        lat1_rad = lon1_rad = -np.Inf
        for stn in allTx+allRx:
            if ('lat' in d['station'][stn].keys()
                and d['station'][stn]['lat'] != ''):
                lat0_rad = min(lat0_rad, d['station'][stn]['lat'])
                lon0_rad = min(lon0_rad, d['station'][stn]['lon'])
                lat1_rad = max(lat1_rad, d['station'][stn]['lat'])
                lon1_rad = max(lon1_rad, d['station'][stn]['lon'])
        latCtr_rad = (lat0_rad + lat1_rad)/2
        lonCtr_rad = (lon0_rad + lon1_rad)/2

    grid = d['coverage']['grid']         # Perform grid x grid link budgets.
    range_m = 1e3*d['coverage']['range'] # Radius of coverage map.

    # Calculate edges of map.
    _,lonE_rad,_ = earth.direct(latCtr_rad, lonCtr_rad, np.pi/2, range_m)
    _,lonW_rad,_ = earth.direct(latCtr_rad, lonCtr_rad, -np.pi/2, range_m)
    latN_rad,_,_ = earth.direct(latCtr_rad, lonCtr_rad, 0, range_m)
    latS_rad,_,_ = earth.direct(latCtr_rad, lonCtr_rad, np.pi, range_m)
    deltaLat_rad = abs(latN_rad - latS_rad)/grid
    deltaLon_rad = abs(lonE_rad - lonW_rad)/grid

    # Noise power.  Potentially, unique noise per receiver - just 1 here!.
    nRxName = stnRx['noise']
    nRx = d['noise'][nRxName]
    enb_Hz = nRx['B']        # Equivalent noise bandwidth of signal.
    env    = nRx['env']      # Receiver environment.
    f_MHz  = float(nRx['f']) # Frequency of signal.
    nf_dB  = nRx['NF']       # Receiver noise figure.
    rxN_dBm = noise.noise_dBm(f_MHz, enb_Hz, nf_dB, env)

    # Longley-Rice ITM to predict signal strength at receiver.
    rxS_dBm = np.zeros((grid, grid)) # Signal power at receiver.
    # Perform link budgets from tx to each square in grid.
    for i in range(grid):
        latRx_rad = latS_rad + (i+0.5)*deltaLat_rad # 0.5 for cell center.
        stnRx['latCell'] = latRx_rad
        for j in range(grid):
            lonRx_rad = lonW_rad + (j+0.5)*deltaLon_rad
            stnRx['lonCell'] = lonRx_rad
            rxS_dBm[i][j],_,_,pAng,pGain,_ = linkBudget(d, stnTxName, stnRxName)
            if not progress is None:
                progress(int(100*(i*grid + j)/(grid*grid)))
    area_rad = [lonW_rad, lonE_rad, latS_rad, latN_rad]

    # Generate command to manually retrieve MapQuest map.
    print('Retrieve map: mapquest %.6f ' % np.degrees(latS_rad), end='')
    print('%.6f ' % np.degrees(lonW_rad), end='')
    print('%.6f ' % np.degrees(latN_rad), end='')
    print('%.6f' % np.degrees(lonE_rad))
    return rxS_dBm, rxN_dBm, area_rad, pAng, pGain

def links(d):
    '''Perform pairwise tx-rx link budgets for all transmitters and
    receivers.  Results are written to a file in csv (comma separated value)
    format.

    Inputs:
      d (dictionary): dictionary of rfl configuration.
    Output:
      none
    '''

    filename = d['rfl']['outfile']
    f = open(filename, 'w')
    hdr = 'txName,rxName,tx_dBm,gainTx_dBi,gainRx_dBi,pl_dB,rx_dBm,n_dBm,mode\n'
    f.write(hdr)
    allTx, allRx = stnGroups(d)
    for stnTx in allTx:
        latTx_rad = d['station'][stnTx]['lat']
        lonTx_rad = d['station'][stnTx]['lon']
        for stnRx in allRx:
            latRx_rad = d['station'][stnRx]['lat']
            lonRx_rad = d['station'][stnRx]['lon']
            s_dBm, gain_dBi, pl_dB,_,_,pm = linkBudget(d, stnTx, stnRx)

            nRxName = d['station'][stnRx]['noise']
            nRx = d['noise'][nRxName]
            enb_Hz = nRx['B']        # Equivalent noise bandwidth of signal.
            env    = nRx['env']      # Receiver environment.
            f_MHz  = float(nRx['f']) # Frequency of signal.
            nf_dB  = nRx['NF']       # Receiver noise figure.
            n_dBm = noise.noise_dBm(f_MHz, enb_Hz, nf_dB, env)
            tx_dBm = d['station'][stnTx]['tx']
            s = ('%s,%s,%.1f,%.1f,%.1f,%.1f,%.1f,%.1f,%s\n' % (stnTx, stnRx,
                tx_dBm, gain_dBi[0], gain_dBi[1], pl_dB, s_dBm, n_dBm, pm))
            f.write(s)
    f.close()

def pathloss_dB(rng, pfl, d, stnTxName, stnRxName):
    '''Use ITM model to predict RF path loss over terrain between two
    stations.

    Inputs:
      rng (float[]): ranges n meters from tx station.
      pfl (float[]): elevations above sea level in meters at each rng[].
      d (dictionary): dictionary of rfl configuration.
      stnTxName (string): name of transmit station.
      stnRxName (string): name of receive station.
    Outputs:
      loss_dB (float): ITM path loss between tx and rx stations.
      elTx_rad (float): RF take off angle of transmit station.
      elRx_rad (float): RF take off angle of receive station.
    '''

    global dhItm, dhItmN

    itm.dictionaries()  # Set ITM values to defaults.

    stnTx = d['station'][stnTxName]
    stnRx = d['station'][stnRxName]
    itm.prop['hg'][0] = stnTx['hgt'] # hg is Height above Ground.
    itm.prop['hg'][1] = stnRx['hgt']

    # Get environmental values needed by ITM.
    elev_m  = (stnTx['hgt'] + stnRx['hgt'])/2 # Avg elev above sea level.
    env     = d['env'][stnTx['env']]
    epsR    = env['eps']       # Relative ground permittivity, F/m.
    nName   = stnTx['noise']   # Name of this station's noise model.
    f_MHz   = float(d['noise'][nName]['f']) # Signal frequency.
    N0      = env['N0']        # Sea level refractivity in given region.
    polAnt  = stnTx['pol']     # Vertical antenna polarization.
    sigma   = env['sigma']     # Relative ground conductivity, Siemens.
    climate = env['climate']   # Receiver climate.
    service = stnTx['service'] # Transmission mode.
    service += 10              # Tell itm.aVar() we're in profile mode.
    if d['rfl']['interference'] == 'yes':
        service += 20

    # Memory check.
    if climate != mem['climate']:
        mem['climate'] = climate
    else:
        climate = -1
    if service != mem['service']:
        mem['service'] = service
    else:
        climate = -1

    # Calculate median path loss.  aVar() follow up calls are recommended.
    itm.lrPrep(f_MHz, elev_m, N0, polAnt, epsR, sigma)
    itm.lrProfile(rng, pfl, climate, service) # Calculate path loss.
    dhItm += itm.prop['dh']
    dhItmN += 1

    # Retrieve answers from ITM dictionary.
    dist_m   = itm.prop['dist']
    fs_dB    = freespace_dB(None, 1e6*f_MHz, dist_m)
    elTx_rad = itm.prop['the'][0] # RF takeoff angle at transmitter.
    elRx_rad = itm.prop['the'][1] # RF takeoff angle at receiver.

    # Apply confidence and reliability levels.
    if 'conf' in d['rfl'].keys() and d['rfl']['conf'] != '':
        qc = d['rfl']['conf']
        if 'time' in d['rfl'].keys():
            qr = d['rfl']['time']
        else:
            qr = d['rfl']['time'] = 50
        zc = norm.isf(qc/100)
        zr = norm.isf(qr/100)
        med_dB, _ = itm.aVar(zr, 0, zc)
    else:
        med_dB = itm.prop['aref']
    loss_dB = fs_dB + med_dB

    return loss_dB, elTx_rad, elRx_rad

def rfPredict(d, gui=False):
    # Output can profile or area mode.  Each mode can be presented in a point
    # to point manner or coverage map.

    rfl = d['rfl']
    if not 'output' in rfl.keys():
        print('rfl: Output specifier not found.  Must be \'coverage\' '
            'or \'cvs\'.')
        sys.exit(1)
    match rfl['output']:
        case 'csv':
            links(d)
        case 'coverage':
            # Calculate link budget to each cell in coverage grid.
            s_dBm, n_dBm, area_rad, pAng, pGain = coverage(d)
            units = rfl['units']
            match units:
                case 'dBuV':
                    # Use sig uV above noise uV.
                    n_uV = dBm_to_uV(n_dBm) # dBm to uV.
                    s_uV = dBm_to_uV(s_dBm) # dBm to uV.
                    sn_uV = s_uV - n_uV
                    i = np.where(sn_uV<=0)
                    sn_uV[i] = 1e-20
                    sig = 20*np.log10(sn_uV) # uV to dBuV.
                case 'uV': # Noise not considered.
                    sig = dBm_to_uV(s_dBm) # dBm to uV.
                case 'dBm': # Noise not considered.
                    sig = s_dBm
                case 'SNR':
                    sig = s_dBm - n_dBm
                case _:
                    print('Unexpected output units %s.' % units)
                    sys.exit(1)
            contour(d, sig, area_rad, pAng, pGain, gui)
        case _:
            print('rfl: unexpected output \'%s\'.  Must be \'coverage\' or '
                '\'cvs\'.' % rfl['output'])
    if rfl['mode'] == 'profile' and rfl['output'] == 'coverage':
        print('From terrain: dh = %d (Area Mode input)'
            % int(round(dhItm/dhItmN)))

def usage():
    u = '''Usage: rfl [-c file] [-d file] [-t title]
           -c file: read configuration from 'file'.
           -d file: write default commented configuration to 'file'.'''
    print(u)

def main(argv):
    d = cli(argv) # Get dictionary of cmd line args.
    if d is None: # No command line args passed, so start GUI.
        g = gui.GuiConnect()
        g.gui()
        sys.exit(0)
    rfPredict(d)

if __name__ == '__main__':
    main(sys.argv)

